/* mgen.c
   microprogram generator
   from microprogram spec.  mspec.txt

   26 July 98
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define	 MXBUF  1024		   /* max length of input line */
#define	 MXMS	128		   /* max size of micro store */
#define	 MXW	64		   /* max width of micro store */

char 	ibuf[MXBUF], *cp = NULL;   /* input buffer, ch ptr */
static char sep[] = " \t\n";       /* separator char */
int 	lineno,loc;		   /* line num, current mpgm ads */
int 	width,ads1,ads2;           /* width of mpgm, bit no. of Goto */
FILE 	*fi;

char 	ms[MXMS][MXW];		   /* micro store */

#include "hash.c"		   /* symbol table routines */

seterr(char *s)
{
  printf("line %d error : %s\n",lineno,s);
  exit(0);
}

char *tok()	/* get one token from input */
{
  if( cp != NULL ) cp = strtok(NULL,sep);
  while ( cp == NULL ) {
    if( fgets(ibuf,MXBUF,fi) == NULL ) return NULL;
    lineno++;
    cp = strtok(ibuf, sep);
  }
  return cp;
}

int iscmd( char *w )	/* Is token a command ? */
{
  return ( w[0] == '.' );
}

int eq( char *s1, char *s2)	/* string equal */
{
  return( strcmp(s1,s2) == 0);
}

dosymbol()	/* get token update symbol table until other cmd */
{
  char *w;
  int v;
  w = tok();
  while ( w != NULL && ! iscmd(w) ) {
    v = atoi(tok());
    putsym(w,v);
    w = tok();
  }
}

dolabel()	/* get label update symbol table until othe cmd */
{
  char *w;
  w = tok();
  while ( w != NULL && ! iscmd(w) ) {
    if( w[0] == ':' ) putsym(w+1,loc);
    if( eq(w,";") ) loc++;
    w = tok();
  }
}

pass1()		/* collect symbols */
{
 char *w;
 w = tok();
 while ( iscmd(w) ) {
   switch( w[1] ) {
   case 'e' : return;
   case 'w' :
     width = atoi(tok());
     w = tok();
     break;
   case 'a' :
     ads1 = atoi(tok());
     ads2 = atoi(tok());
     w = tok();
     break;
   case 's' : dosymbol(); break;
   case 'm' : dolabel(); break;
   }
 }
}

char getbit(int i, int n)
{
  if( (n >> i) & 1 ) return '1';
  return '0';
}

/*
    m[loc][ads]   = getbit(4,a);
    m[loc][ads+1] = getbit(3,a);
    ...
    m[loc][ads+5] = getbit(0,a);
*/
putads(int a)
{
  int i;
  for( i=ads1; i<=ads2; i++ )
    ms[loc][i] = getbit(ads2-i,a);
}

pass2()		/* generate microprogram */
{
  char *w;
  int a;
  w = tok();
  while ( w != NULL ) {		/* skip to .m section */
    if( eq(w,".e")) return;
    if( eq(w,".m")) break;
    w = tok();
  }
  w = tok();
  while ( w != NULL ) {
    if( eq(w,".e")) return;
    switch( w[0] ) {
    case ':' : break;
    case ';' : loc++; break;
    case '/' :
      a = getsym(w+1);
      if ( a < 0 ) seterr("symbol not found");
      putads(a);
      break;
    default :
      a = getsym(w);
      if ( a < 0 ) seterr("symbol not found");
      ms[loc][a] = '1';
      break;
    }
    w = tok();
  }
}

printout(int n)
{
  int i,j;
  printf("%d\n",n);
  for( i=0; i<n; i++ ) {
    printf("%3d ",i);
    for( j=0; j<width; j++ )
      printf("%c",ms[i][j]);
    printf("\n");
  }
}

initms(int height, int width)
{
  int i,j;
  for( i=0; i<height; i++ )
    for( j=0; j<width; j++ )
      ms[i][j] = '0';
}

main()
{
  fi = fopen("mspec.txt","r");
  lineno = 0;
  loc = 0;
  pass1();
  rewind(fi);
  initms(loc,width);
  lineno = 0;
  loc = 0;
  pass2();
  fclose(fi);
  printout(loc);
}
